/*
 * Copyright (c) 2003-2005 Erez Zadok
 * Copyright (c) 2003-2005 David P. Quigley
 * Copyright (c) 2003-2005 Charles P. Wright
 * Copyright (c) 2003-2005 Stony Brook University
 * Copyright (c) 2003-2005 The Research Foundation of State University of New York
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package.
 *
 * This Copyright notice must be kept intact and distributed with all sources.
 */
/*
 *  $Id: unionimap.c,v 1.5 2005/06/16 17:56:21 dquigley Exp $
 */

#include "unionimap.h"

/**
 *	print_usage()
 *	Function used to print out usage information 
 *	when an error is encountered
 */
void print_usage()
{
		fprintf(stderr, "unionimap version: $Id: unionimap.c,v 1.5 2005/06/16 17:56:21 dquigley Exp $\n");
		fprintf(stderr, "Distributed with Unionfs "UNIONFS_VERSION"\n");
		fprintf(stderr, "\n");
		fprintf(stderr,"Usage: unionfsimap [-c] [-a ARG:] filename path\n");
		fprintf(stderr," flag -c: create forward map with name filename\n");
		fprintf(stderr," flag -a ARG: create reverse map with name filename and path and add it to forwardmap ARG\n");
		fprintf(stderr," flag -d: if set once will dump the header of the map if twice it will also dump the contents\n");
		fprintf(stderr," flag -h: prints this message then quits\n");
		fprintf(stderr," please note that -c -a and -d are mutually exclusive\n");
}

/**
 * create_forwardmap(char *filename)
 * verifies that the forward map is valid.
 *
 */
int create_forwardmap(char *filename)
{
		int byteswritten = 0,err = 0;
		FILE *fwrdmap = NULL;
		struct fmaphdr header;
		void *table = NULL;
		uuid_t uuid;

		fwrdmap = fopen(filename,"w");
		if(!fwrdmap){
				err =  -EINVAL;
				goto out;
		}
		header.version = FORWARDMAP_VERSION;
		header.magic = FORWARDMAP_MAGIC;
		header.usedbranches = 0;
		uuid_generate(uuid);
		memcpy(&header.uuid,&uuid,UUID_LEN);
		byteswritten = fwrite((void *)&header,sizeof(struct fmaphdr),1,fwrdmap);
		if(byteswritten < sizeof(struct fmaphdr)){
			fprintf(stdout,"BytesWritten for header:%d",byteswritten);	
				unlink(filename);
				err = -EIO;
				perror("Failed Writting header: ");
				goto out;
		}	
		table = malloc(sizeof(uuid_t[256]));
		if(!table){
				err = -ENOMEM;
				goto out;
		}
		memset(table,0,sizeof(uuid_t[256]));
		byteswritten = fwrite(table,sizeof(uuid_t[256]),1,fwrdmap);
		if(byteswritten < sizeof(uuid_t[256])){
			fprintf(stdout,"BytesWritten for header:%d",byteswritten);
			unlink(filename);
				err = -EIO;
				perror("Failed writting table: ");
				
				goto out;
		}
out:
		if(fwrdmap){
				fclose(fwrdmap);
		}
		if(table){
				free(table);
		}

		return err;
}

int create_reversemap(char *filename, char* path, char* forwardmap)
{
		int byteswritten = 0,bytesread = 0,err = 0;
		off_t offset = 0;
		FILE *fwrdmap = NULL,*revmap = NULL;
		struct fmaphdr fwdheader;
		struct rmaphdr revheader;
		struct statfs stat;
		uuid_t uuid;

		fwrdmap = fopen(forwardmap,"r");
		if(!fwrdmap){

				err =  -EINVAL;
				perror("Open on Forwardmap  Failed: ");
				goto out;
		}	
		bytesread = fread((void*)&fwdheader,sizeof(struct fmaphdr),1,fwrdmap);
		if(bytesread < sizeof(struct fmaphdr)){
				err =  -EINVAL;
				goto out;
		}
		if(fwdheader.magic != FORWARDMAP_MAGIC || fwdheader.version != FORWARDMAP_VERSION){
				err = -EINVAL;
				goto out;
		}
		if(fwdheader.usedbranches >= 256){
				fprintf(stderr,"Forwardmap already contains maximum number of reverse maps");
				err = -EINVAL;
				goto out;
		}
		memset(&stat,0,sizeof(struct statfs));
		err = statfs(path,&stat);
		if(err){
				perror("statfs failed on path: ");
				goto out;
		}
		revmap = fopen(filename,"w");
		if(!revmap){
				err = -EINVAL;
				perror("Open on Reversemap failed: ");
				goto out;
		}
		revheader.version = REVERSEMAP_VERSION;
		revheader.magic = REVERSEMAP_MAGIC;
		uuid_generate(uuid);
		memcpy(&revheader.revuuid,&uuid,UUID_LEN);
		memcpy(&revheader.fwduuid,&fwdheader.uuid,UUID_LEN);
		revheader.fsid = stat.f_fsid;
		byteswritten = fwrite(&revheader,sizeof(struct rmaphdr),1,revmap);
		if(byteswritten < sizeof(struct rmaphdr)){
				unlink(filename);
				err = -EIO;
				perror("Reversemap Write failed: ");
				goto out;
		}

		offset = sizeof(struct fmaphdr) + (fwdheader.usedbranches * sizeof(uuid_t));
		err = fseek(fwrdmap,offset,SEEK_SET);
		if(err){
				unlink(filename);
				err = -EIO;
				perror("Couldent seek to offset in uuid table: ");
		}
		byteswritten = fwrite(&revheader.fwduuid,sizeof(uuid_t),1,fwrdmap);
		if(byteswritten < sizeof(uuid_t)){
				unlink(filename);
				err = -EIO;
				perror("Forwardmap Write failed to write uuid to table: ");
				goto out;
		}
		fwdheader.usedbranches++;
		offset = (int)&(((struct fmaphdr *)(0))->usedbranches);
		err = fseek(fwrdmap,offset,SEEK_SET);
		if(err){
				unlink(filename);
				err = -EIO;
				perror("Couldent seek to offset in uuid table: ");
		}
		byteswritten = fwrite(&fwdheader.usedbranches,sizeof(uint8_t),1,fwrdmap);
		if(byteswritten < sizeof(uint8_t)){
				unlink(filename);
				err = -EIO;
				perror("Forwardmap Write failed to update usedbranches: ");
				goto out;
		}	
out:
		if(fwrdmap){
				fclose(fwrdmap);
		}
		if(revmap){
				fclose(revmap);
		}
		return err;
}
int print_forwardmap(FILE *file,int debug_level)
{
		int err = 0, bytesread = 0,i;
		struct fmaphdr header;
		uuid_t btable[256];
		char *uuid_unparsed;
		struct fmapent entry;

		if(!file || debug_level <= 0){
				err = -EINVAL;
				goto out;
		}	
		err = fseek(file,0L,SEEK_SET);
		if(err){
				goto out;
		}
		bytesread = fread((void *)&header,sizeof(struct fmaphdr),1,file);
		if(bytesread != sizeof(struct fmaphdr)){
				err = -EIO;
				goto out;
		}
		uuid_unparsed = (char *)malloc(37);
		if(!uuid_unparsed){
				err = -EIO;
				goto out;
		}
		memset(uuid_unparsed,0,37);
		fprintf(stdout,"Unionfs Forwardmap:\n");
		fprintf(stdout,"Magic: %d\n",header.magic);
		fprintf(stdout,"Version: %d\n",header.version);
		fprintf(stdout,"Used Branches: %d\n",header.usedbranches);
		uuid_unparse(header.uuid,uuid_unparsed);
		fprintf(stdout,"UUID: %s\n",uuid_unparsed);
		bytesread = fread((void *)&btable,sizeof(uuid_t[256]),1,file);
		if(bytesread != sizeof(struct fmaphdr)){
				err = -EIO;
				goto out;
		}
		for(i = 0; i < header.usedbranches;i++){
				fprintf(stdout,"Branch at index : %d\n",i);
				uuid_unparse(btable[i],uuid_unparsed);
				fprintf(stdout,"uuid: %s\n",uuid_unparsed);
		}		
		if(debug_level > 1){
				fprintf(stdout,"File Contents\n");
				while(fread((void *)&entry,sizeof(struct fmapent),1,file)){
						fprintf(stdout,"fsnum: %d inode: %llu\n",(uint8_t)entry.fsnum,entry.inode);
				}
		}
out:
		return err;

}
int print_reversemap(FILE* file,int debug_level)
{

		int err = 0, bytesread = 0;
		struct rmaphdr header;
		char *uuid_unparsed = NULL;
		uint64_t inode;

		if(!file || debug_level <= 0){
				err = -EINVAL;
				goto out;
		}	
		err = fseek(file,0L,SEEK_SET);
		if(err){
				goto out;
		}
		bytesread = fread((void *)&header,sizeof(struct fmaphdr),1,file);
		if(bytesread != sizeof(struct fmaphdr)){
				err = -EIO;
				goto out;
		}
		uuid_unparsed = (char *)malloc(37);
		if(!uuid_unparsed){
				err = -ENOMEM;
				goto out;
		}
		memset(uuid_unparsed,0,37);
		fprintf(stdout,"Unionfs Reversemap:\n");
		fprintf(stdout,"Magic: %d\n",header.magic);
		fprintf(stdout,"Version: %d\n",header.version);
		uuid_unparse(header.fwduuid,uuid_unparsed);
		fprintf(stdout,"Forward UUID: %s\n",uuid_unparsed);
		uuid_unparse(header.revuuid,uuid_unparsed);
		fprintf(stdout,"Reverse UUID: %s\n",uuid_unparsed);
		fprintf(stdout,"fsid: %04x%04x",((unsigned int *)&header.fsid)[0],((unsigned int *)&header.fsid)[1]);
		if(debug_level > 1){
				fprintf(stdout,"File Contents\n");
				while(fread((void *)&inode ,sizeof(uint64_t),1,file)){
						if(inode){

								fprintf(stdout,"Pos: %lu , ino: %llu\n",ftell(file),inode);
						}
				}
		}
out:
		if(uuid_unparsed){
				free(uuid_unparsed);
		}
		return err;
}
int dump_map(char *filename, int debug_level)
{
		int err = 0,bytesread = 0;
		uint32_t magic;
		FILE *file = NULL;

		file = fopen(filename,"r");
		if(!file){
				err = -EINVAL;
				goto out;
		}
		bytesread = fread((void *)&magic,sizeof(uint32_t),1,file);
		if(bytesread < sizeof(uint32_t)){
				err = -EINVAL;
				goto out;
		}
		if(magic == FORWARDMAP_MAGIC){
				err = print_forwardmap(file,debug_level);
		} else if(magic == REVERSEMAP_MAGIC) {
				err = print_reversemap(file,debug_level);
		}
		else{
				err = -EINVAL;
		}
out:
		if(file){
				fclose(file);
		}

		return err;
}

int main(int argc, char** argv)
{
		int c = 0; //Temp varliable to hold our character for the switch
		int cflag = 0,aflag = 0,dflag = 0;
		int err = 0;
		char *avalue = NULL;
		if(argc < 2)
		{
				print_usage();
				err = -EINVAL;
				goto out;
		}
		while ((c = getopt (argc, argv, "a:cd")) != -1)
		{
				switch(c)
				{
						case 'c':
								cflag = c;
								break;
						case 'a':
								aflag = 1;
								avalue = (char*)malloc(sizeof(char)*(strlen(optarg)+1));
								strcpy(avalue,optarg);
								break;
						case 'd':
								dflag++;
								break;
						case '?':
								if(isprint (optopt))
										fprintf (stderr, "Unknown option `-%c'.\n", optopt);
								else
										fprintf (stderr,"Unknown option character `\\x%x'.\n",optopt);
								print_usage();
								err = -EINVAL;
								goto out;
								break;
						default:
								print_usage();
								err = -EINVAL;
								goto out;
								break;	
				}
		}
		if(optind == argc){
				print_usage();
				err = -EINVAL;
				goto out;
		}
		if(!(cflag^aflag^dflag)){
				print_usage();
				err = -EINVAL;
				goto out;
		}
		if(cflag){
				err = create_forwardmap(argv[optind]);
				if(err){
						unlink(argv[optind]);
				}
				goto out;
		} else if(aflag){
				if(optind+1 > argc){
						print_usage();
						err = -EINVAL;
						goto out;
				}

				err = create_reversemap(argv[optind],argv[optind+1],avalue);
				if(err){
						unlink(argv[optind]);
						goto out;
				}

		} else if(dflag){
				err = dump_map(argv[optind],dflag);
				if(err){
						goto out;
				}
		}
out:
		if(avalue){
				free(avalue);
		}
		return err;
}

/*
 * vim:shiftwidth=8
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
